/**
 * class KImage
 * A smart-scaling Image wrapper
 *  Allows for an image to be specified as a group of images arranged in a
 *  pattern, and will only scale certain images in the pattern.
 *
 * Copyright 2005, James M. Allen
 * You are free to reuse this code, provided you keep this and the above
 * copyright notice intact.
 *
 * Requirements
 *  - Yahoo! Widget Engine v3.0 or higher
 *  - A transparent 1x1 "blank" image found at Resources/Blank.png
 *
 * Usage example:
      
      include("KImage.js");
      foo = new KImage();
      foo.src = "Resources/Window*.png";
      foo.hOffset = 30;
      foo.vOffset = 30;
      foo.width = 200;
      foo.height = 150;
      
      var anm = KResizeAnimation(foo, 400, 300, animator.kEaseOut, doneFunc);
      animator.start(anm);
  
 *  The above example will create a KImage called foo, that is made up of a
 *  3x3 grid of images (WindowTopLeft.png, WindowTopCenter.png, ...).
 *  It is initially set to a size of (200, 150), and an animation is started
 *  to resize it to (400, 300). Because the image is a 3x3 grid (specified by
 *  the "*" character in the src), the images at the corners will maintain
 *  their original sizes, and the other images will stretch to fill the rest
 *  of the width and the height.
 *
 * Details
 *  The key to specifying a pattern of images is in the src property.
 *  If src contains one of the three special characters (*, <, or |), then
 *  the image will take on a pattern according to the character and look
 *  for images with matching filenames. The patterns are as follows:
 *
 *  3x3 (Specified with "*")
 *    This is a 3x3 pattern. Images in the corners (TopLeft, TopRight,
 *    BottomLeft, and BottomRight) will maintain their original size. Images
 *    on the left and right will maintain their original width, and images
 *    on the top and bottom will maintain their original height.
 *    
 *    If src is specified as "MyImage*.png", you need the following files:
 *    MyImageTopLeft.png      MyImageTopCenter.png      MyImageTopRight.png
 *    MyImageMiddleLeft.png   MyImageMiddleCenter.png   MyImageMiddleRight.png
 *    MyImageBottomLeft.png   MyImageBottomCenter.png   MyImageBottomRight.png
 *
 *  Refer to example above or the function definitions below for more
 *  information.
 *
 *
 */


/**
 * class KImage
 *  An Image object replacement.
 *  Should support most of the properties that a normal Image supports,
 *  with added functionality of "smart-scaling" images.
 */
function KImage(parentWindow)
{
  this.id = KImage.ids.length;
  KImage.ids.push(this);
  
  // PRIVATE VARIABLES
  var frame = new Frame();
  var img = new Array();
  for (var i = 0; i < 9; i++) {
    var curI = new Image(parentWindow);
    img.push(curI);
    curI.visible = false;
    curI.src = "Resources/Blank.png";
    curI.width = -1;
    curI.height = -1;
    frame.addSubview(curI);
  }
  var pattern = null;
  
  var self = this;
  
  // PRIVATE METHODS
  // Privileged proxy
  var setSrcProxy = function(prop, oldval, newval){ return this.setSrc(prop, oldval, newval); }
  this.setSrc = function(prop, oldval, newval){     return setSrc(prop, oldval, newval); }
  var setSrc = function(prop, oldval, newval)
  {
    for (var i in img) {
      img[i].visible = false;
      img[i].src = "Resources/Blank.png";
      // img[i].width = -1;
      // img[i].height = -1;
    }
    
    if (!newval || typeof(newval) != "string") {
      pattern = "unknown";
      self.watch("minHeight", setReadWrite);
      self.watch("minWidth", setReadWrite);
      self.watch("srcHeight", setReadWrite)
      self.watch("srcWidth", setReadWrite);
      self.minWidth = 0;
      self.minHeight = 0;
      self.srcWidth = 0;
      self.srcHeight = 0;
      self.watch("minHeight", setReadOnly);
      self.watch("minWidth", setReadOnly);
      self.watch("srcHeight", setReadOnly);
      self.watch("srcWidth", setReadOnly);
      return "";
    }
    
    log("pre-indexOf {");
    
    if (newval.indexOf("*") != -1) {
      // log("  pattern = 3x3");
      pattern = "3x3";
      img[KImage.kTopLeft].src = newval.replace("*", "TopLeft");
      img[KImage.kTopLeft].fillMode = "tile";
  img[KImage.kTopLeft].onMouseDown  = 'sg_activateGrip();';
	img[KImage.kTopLeft].onMouseMove  = 'sg_dragGrip("NW");';
  img[KImage.kTopLeft].onMouseUp    = 'dragObjectRelease(grippedObject);';
      img[KImage.kTopCenter].src = newval.replace("*", "TopCenter");
      img[KImage.kTopCenter].fillMode = "tile";
  //img[KImage.kTopCenter].onMouseDown  = 'sg_activateGrip();';
	//img[KImage.kTopCenter].onMouseMove  = 'sg_dragGrip("N");';
  //img[KImage.kTopCenter].onMouseUp    = 'dragObjectRelease(grippedObject);';
  img[KImage.kTopCenter].onMouseDown  = 'updateGrips(' + parentWindow.name + ');';
	img[KImage.kTopCenter].onMouseMove  = 'dragObject(' + parentWindow.name + ');';
  img[KImage.kTopCenter].onMouseUp    = 'dragObjectRelease(' + parentWindow.name + ');';
      img[KImage.kTopRight].src = newval.replace("*", "TopRight");
  img[KImage.kTopRight].onMouseDown  = 'sg_activateGrip();';
	img[KImage.kTopRight].onMouseMove  = 'sg_dragGrip("NE");';
  img[KImage.kTopRight].onMouseUp    = 'dragObjectRelease(grippedObject);';
      img[KImage.kMiddleLeft].src = newval.replace("*", "MiddleLeft");
      img[KImage.kMiddleLeft].fillMode = "tile";
  img[KImage.kMiddleLeft].onMouseDown  = 'sg_activateGrip();';
	img[KImage.kMiddleLeft].onMouseMove  = 'sg_dragGrip("W");';
  img[KImage.kMiddleLeft].onMouseUp    = 'dragObjectRelease(grippedObject);';
      img[KImage.kMiddleCenter].src = newval.replace("*", "MiddleCenter");
      img[KImage.kMiddleCenter].opacity = 255;
  img[KImage.kMiddleCenter].onMouseDown  = 'updateGrips(' + parentWindow.name + ');';
	img[KImage.kMiddleCenter].onMouseMove  = 'dragObject(' + parentWindow.name + ');';
  img[KImage.kMiddleCenter].onMouseUp    = 'dragObjectRelease(' + parentWindow.name + ');';
      img[KImage.kMiddleRight].src = newval.replace("*", "MiddleRight");
      img[KImage.kMiddleRight].fillMode = "tile";
  img[KImage.kMiddleRight].onMouseDown  = 'sg_activateGrip();';
	img[KImage.kMiddleRight].onMouseMove  = 'sg_dragGrip("E");';
  img[KImage.kMiddleRight].onMouseUp    = 'dragObjectRelease(grippedObject);';
      img[KImage.kBottomLeft].src = newval.replace("*", "BottomLeft");
  img[KImage.kBottomLeft].onMouseDown  = 'sg_activateGrip();';
	img[KImage.kBottomLeft].onMouseMove  = 'sg_dragGrip("SW");';
  img[KImage.kBottomLeft].onMouseUp    = 'dragObjectRelease(grippedObject);';
      img[KImage.kBottomCenter].src = newval.replace("*", "BottomCenter");
      img[KImage.kBottomCenter].fillMode = "tile";
  img[KImage.kBottomCenter].onMouseDown  = 'sg_activateGrip();';
	img[KImage.kBottomCenter].onMouseMove  = 'sg_dragGrip("S");';
  img[KImage.kBottomCenter].onMouseUp    = 'dragObjectRelease(grippedObject);';
      img[KImage.kBottomRight].src = newval.replace("*", "BottomRight");
  img[KImage.kBottomRight].onMouseDown  = 'sg_activateGrip();';
	img[KImage.kBottomRight].onMouseMove  = 'sg_dragGrip("SE");';
  img[KImage.kBottomRight].onMouseUp    = 'dragObjectRelease(grippedObject);';
      for (var i = 8; i >= 0; --i) {
        img[i].visible = true;
        img[i].width = -1;
        img[i].height = -1;
      }
    }
    // STUFF DELETED BY CRCCHECK
    
    log("} post-indexOf");
    
    
    /*
    log("  img dump");
    for (var i in img) {
      self.idump(img[i]);
    }
    */
    
    self.watch("minHeight", setReadWrite);
    self.watch("minWidth", setReadWrite);
    self.watch("srcHeight", setReadWrite)
    self.watch("srcWidth", setReadWrite);
    
    self.minWidth  = 0;
    self.minHeight = 0;
    
    self.minHeight = img[KImage.kTopLeft].srcHeight + img[KImage.kBottomLeft].srcHeight;
    self.minWidth  = img[KImage.kTopLeft].srcWidth + img[KImage.kTopRight].srcWidth;
    self.srcWidth  = self.minWidth + img[KImage.kTopCenter].srcWidth;
    self.srcHeight = self.minHeight + img[KImage.kMiddleLeft].srcHeight;

    self.watch("minHeight", setReadOnly);
    self.watch("minWidth", setReadOnly);
    self.watch("srcHeight", setReadOnly);
    self.watch("srcWidth", setReadOnly);
    
    
    if (self.width < 0) { self.width = self.srcWidth; }
    if (self.height < 0) { self.height = self.srcHeight; }
    
    log("pre-resize {");
    
    resize(false);
    
    log("} post-resize");
    
    return newval;
  }
  
  
  
  var resize = function(cheap)
  {
    var tHeight;
    var tWidth;
    if (self.height < 0 || !self.height) {
      tHeight = self.srcHeight;
    } else {
      tHeight = self.height;
    }
    if (self.width < 0) {
      tWidth = self.srcWidth;
    } else {
      tWidth = self.width;
    }
        if (!cheap) {
          img[KImage.kTopLeft].hOffset =
            img[KImage.kMiddleLeft].hOffset = 
            img[KImage.kBottomLeft].hOffset = 0;
          img[KImage.kTopCenter].hOffset = 
            img[KImage.kMiddleCenter].hOffset = 
            img[KImage.kBottomCenter].hOffset = img[KImage.kTopLeft].width;
          img[KImage.kTopLeft].vOffset = 
            img[KImage.kTopCenter].vOffset = 
            img[KImage.kTopRight].vOffset = 0;
          img[KImage.kMiddleLeft].vOffset = 
            img[KImage.kMiddleCenter].vOffset = 
            img[KImage.kMiddleRight].vOffset = img[KImage.kTopLeft].height;
        }
        img[KImage.kTopCenter].width = 
          img[KImage.kMiddleCenter].width =
          img[KImage.kBottomCenter].width = tWidth - self.minWidth;
        img[KImage.kTopRight].hOffset = 
          img[KImage.kMiddleRight].hOffset = 
          img[KImage.kBottomRight].hOffset = img[KImage.kTopCenter].hOffset + img[KImage.kTopCenter].width;
        img[KImage.kMiddleLeft].height = 
          img[KImage.kMiddleCenter].height =
          img[KImage.kMiddleRight].height = tHeight - self.minHeight;
        img[KImage.kBottomLeft].vOffset = 
          img[KImage.kBottomCenter].vOffset = 
          img[KImage.kBottomRight].vOffset = img[KImage.kMiddleLeft].vOffset + img[KImage.kMiddleLeft].height;
  }
  
  // Setters
  var setAlignmentProxy = function(prop, oldval, newval){    return this.setAlignment(prop, oldval, newval);  }
  this.setAlignment = function(prop, oldval, newval){    return setAlignment(prop, oldval, newval); }
  var setAlignment = function(prop, oldval, newval)  {    return self.setToFrame("hAlign", oldval, newval)  }
  var setScaleProxy = function(prop, oldval, newval)  {    return this.setScale(prop, oldval, newval);  }
  this.setScale = function(prop, oldval, newval)  {    return setScale(prop, oldval, newval);  }
  var setScale = function(prop, oldval, newval)
  {
    if (newval < 0) {
      newval = 1;
    }
    
    this.scale = newval;
    
    switch (pattern) {
      case "3x3":
        img[KImage.kTopLeft].width = this.scale * img[KImage.kTopLeft].srcWidth;
        img[KImage.kTopRight].width = this.scale * img[KImage.kTopRight].srcWidth;
        img[KImage.kMiddleLeft].width = this.scale * img[KImage.kMiddleLeft].srcWidth;
        img[KImage.kMiddleRight].width = this.scale * img[KImage.kMiddleRight].srcWidth;
        img[KImage.kBottomLeft].width = this.scale * img[KImage.kBottomLeft].srcWidth;
        img[KImage.kBottomRight].width = this.scale * img[KImage.kBottomRight].srcWidth;
        
        img[KImage.kTopLeft].height = this.scale * img[KImage.kTopLeft].srcHeight;
        img[KImage.kTopCenter].height = this.scale * img[KImage.kTopCenter].srcHeight;
        img[KImage.kTopRight].height = this.scale * img[KImage.kTopRight].srcHeight;
        img[KImage.kBottomLeft].height = this.scale * img[KImage.kBottomLeft].srcHeight;
        img[KImage.kBottomCenter].height = this.scale * img[KImage.kBottomCenter].srcHeight;
        img[KImage.kBottomRight].height = this.scale * img[KImage.kBottomRight].srcHeight;
        break;
      case "3x1":
        img[KImage.kLeft].width = this.scale * img[KImage.kLeft].srcWidth;
        img[KImage.kRight].width = this.scale * img[KImage.kRight].srcWidth;
        break;
      case "1x3":
        img[KImage.kTop].height = this.scale * img[KImage.kTop].srcHeight;
        img[KImage.kBottom].height = this.scale * img[KImage.kBottom].srcHeight;
        break;
    }
    
    resize();
  }
  
  
  var setHeightProxy = function(prop, oldval, newval)  {    return this.setHeight(prop, oldval, newval);  }
  this.setHeight = function(prop, oldval, newval)  {    return setHeight(prop, oldval, newval);  }
  var setHeight = function(prop, oldval, newval)
  {
    // log("setHeight(" + newval + ")");
    if (newval < 0) {
      newval = self.srcHeight;
    }
    if (newval == oldval) {
      return newval;
    } else {
      // Height needs to be set before calling resize()
      self.height = Math.max(newval, self.minHeight);
      resize(true);
      return self.height;
    }
  }
  
  var setWidthProxy = function(prop, oldval, newval)  {    return this.setWidth(prop, oldval, newval);  }
  this.setWidth = function(prop, oldval, newval)  {    return setWidth(prop, oldval, newval);  }
  var setWidth = function(prop, oldval, newval)
  {
    // log("setWidth(" + newval + ")");
    if (newval < 0) {
      newval = self.srcWidth;
    }
    if (newval == oldval) {
      return newval;
    } else {
      // Width needs to be set before calling resize()
      self.width = Math.max(newval, self.minWidth);
      resize(true);
      return self.width;
    }
  }
  
  
  /**
   * setToImages
   *  Special-case watcher for properties that should be applied to all images.
   */
  var setToImagesProxy = function(prop, oldval, newval)  {    return this.setToImages(prop, oldval, newval);  }
  this.setToImages = function(prop, oldval, newval)  {    return setToImages(prop, oldval, newval);  }
  var setToImages = function(prop, oldval, newval)
  {
    try {
      // log("setToImages(" + prop + ", " + newval + ")");
      for (var i in img) {
        img[i][prop] = newval;
      }
      // this[prop] is set implicitly
      return newval;
    } catch (ex) {
      log("EXCEPTION");
      pdump(ex);
    }
  }
  
  /**
   * setToFrame
   *  Special-case watcher for properties that should be applied to frame.
   */
  var setToFrameProxy = function(prop, oldval, newval)  {    return this.setToFrame(prop, oldval, newval);  }
  this.setToFrame = function(prop, oldval, newval)  {    return setToFrame(prop, oldval, newval);  }
  var setToFrame = function(prop, oldval, newval)
  {
    // log("setToFrame(" + prop + ", " + newval + ")");
    frame[prop] = newval;
    // this[prop] is set implicitly
    return newval;
  }
  
  /**
   * setReadOnly
   *  Special-case watcher for any properties that are "read-only".
   */
  var setReadOnly = function(prop, oldval, newval)  {    return oldval;  }
  
  /**
   * setReadWrite
   *  Special-case watcher for any properties that are "read/write".
   */
  var setReadWrite = function(prop, oldval, newval)  {    return newval;  }
  
  
  // WATCHERS
  // READ-ONLY PUBLIC PROPERTIES
  this.srcHeight = 0;
  this.srcWidth = 0;
  this.minWidth = 0;
  this.minHeight = 0;
  
  this.watch("minHeight", setReadOnly);
  this.watch("minWidth", setReadOnly);
  this.watch("srcHeight", setReadOnly);
  this.watch("srcWidth", setReadOnly);
  
  
  
  // PUBLIC PROPERTIES
  this.alignment = "left";
  // this.clipRect = null;
  this.colorize = null;
  this.contextMenuItems = null;
  this.fillMode = "stretch";
  this.height = -1;
  this.hAlign = "left";
  this.hOffset = 0;
  // this.hRegistrationPoint = null;
  this.hslAdjustment = null;
  this.hslTinting = null;
  this.name = "KImage" + this.id;
  this.onContextMenu = "";
  this.onDragDrop = "";
  this.onDragEnter = "";
  this.onDragExit = "";
  this.onMouseDown = "";
  this.onMouseEnter = "";
  this.onMouseExit = "";
  this.onMouseMove = "";
  this.onMouseUp = "";
  this.onMouseWheel = "";
  this.onMultiClick = "";
  this.opacity = 255;
  // this.rotation = 0;
  this.scale = 1.0;
  this.src = null;
  this.tileOrigin = "topLeft";
  this.tooltip = "";
  this.tracking = "opacity";
  // this.useFileIcon = false;
  this.visible = true;
  this.vOffset = 0;
  this.vAlign = "top";
  // this.vRegistrationPoint = null;
  this.width = -1;
  
  this.watch("alignment", setAlignmentProxy);
  this.watch("colorize", setToImagesProxy);
  this.watch("contextMenuItems", setToImagesProxy);
  this.watch("fillMode", setToImagesProxy);
  this.watch("height", setHeightProxy);
  this.watch("hAlign", setToFrameProxy);
  this.watch("hOffset", setToFrameProxy);
  this.watch("hslAdjustment", setToImagesProxy);
  this.watch("hslTinting", setToImagesProxy);
  this.watch("onContextMenu", setToFrameProxy);
  this.watch("onDragDrop", setToFrameProxy);
  this.watch("onDragEnter", setToFrameProxy);
  this.watch("onDragExit", setToFrameProxy);
  this.watch("onMouseDown", setToFrameProxy);
  this.watch("onMouseEnter", setToFrameProxy);
  this.watch("onMouseExit", setToFrameProxy);
  this.watch("onMouseMove", setToFrameProxy);
  this.watch("onMouseUp", setToFrameProxy);
  this.watch("onMouseWheel", setToFrameProxy);
  this.watch("onMultiClick", setToFrameProxy);
  this.watch("opacity", setToFrameProxy);
  this.watch("scale", setScaleProxy);
  this.watch("src", setSrcProxy);
  this.watch("tileOrigin", setToImagesProxy);
  this.watch("tooltip", setToImagesProxy);
  this.watch("visible", setToFrameProxy);
  this.watch("vOffset", setToFrameProxy);
  this.watch("vAlign", setToFrameProxy);
  this.watch("width", setWidthProxy);
  this.watch("window", setToFrameProxy);
  this.watch("zOrder", setToFrameProxy);
  
  // Create a dummy image to get default values in current context
  var placerImage = new Image(parentWindow);
  this.zOrder = placerImage.zOrder;
  this.window = placerImage.window;
  placerImage = null;
  
}


/**
 * static KImage.ids
 *  Array of all KImage objects - used for event-handling.
 */
KImage.ids = new Array();

/**
 * KImage constants
 *  Constants for array indices of images in particular positions.
 */
KImage.kTop = 0;
KImage.kMiddle = 1;
KImage.kBottom = 2;

KImage.kLeft = 0;
KImage.kCenter = 1;
KImage.kRight = 2;

KImage.kTopLeft = 0;
KImage.kTopCenter = 1;
KImage.kTopRight = 2;
KImage.kMiddleLeft = 3;
KImage.kMiddleCenter = 4;
KImage.kMiddleRight = 5;
KImage.kBottomLeft = 6;
KImage.kBottomCenter = 7;
KImage.kBottomRight = 8;



